# Standard library modules.
import abc
import sys
import queue
import logging
from functools import wraps
from time import time, sleep
from threading import Lock, RLock
from contextlib import contextmanager

# Third party modules.
from tqdm.autonotebook import tqdm

# Local modules

# Globals and constants variables.
FORMAT_STRING = '{asctime}  {threadName:<12}  {levelname:>8}:  {message}'

handler = logging.StreamHandler(stream=sys.stdout)
handler.setFormatter(logging.Formatter(FORMAT_STRING, style='{'))

logger = logging.getLogger('mosquito')
logger.addHandler(handler)


class MosquitoError(Exception):
    """mosquito specific error"""


def args_to_kwargs(_func=None, ftype='function', keep=0):
    """
    Decorator for functions and (static, class-) methods that converts positional arguments into key
    word arguments.

    :param _func: function or method to decorate
    :param ftype: type of function, one of: `function`, `static`, `method` and `classmethod`
    :type ftype: string
    :param keep: number of positional arguments that will be forwarded rather than evaluated.
    :type keep: int
    :return: decorator or a wrapper if function is specified
    """
    kw_func_like = {'function', 'static'}
    kw_meth_like = {'method', 'classmethod'}
    kw_all = {*kw_func_like, *kw_meth_like}

    # validate function type
    assert ftype in kw_all, f'`method` has to be one of {kw_all}!'

    # compute offset for positional arguments to be evaluated, default is 0 (evaluate all)
    offset = (1 if ftype in kw_meth_like else 0) + keep

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # split args in those that are kept and those, that are used to update kwargs
            keep = args[:offset]
            args = args[offset:]

            # update kwarg args with what was found in
            for i in range(len(args)):
                arg_dict = dict(args[-i - 1])
                arg_dict.update(kwargs)
                kwargs = arg_dict

            return func(*keep, **kwargs)

        return wrapper

    if _func:
        return decorator(_func)

    return decorator


def evaluate(value, *args, **kwargs):
    """
    This is a helper function used for lazy evaluation.

        1. If value is a callable it's called with given parameters
        2. If value is an iterable is converted to a tuple

    :param value: value to evluate
    :param args: positional arguments that are passed if value is a callable
    :param kwargs: key word arguments that are passed if value is a callable
    :return: evaluate value
    """
    if callable(value):
        value = value(*args, **kwargs)

    try:
        return tuple(value)

    except TypeError:
        return value


def monitor_queue(qclass):
    """
    Decorator that wraps any queue class in order to observer its progress using tqdm.

    :param qclass: queue class
    :return: monitored wrapper for given queue class
    """
    class _MonitoredQueue(qclass):
        @wraps(qclass.__init__)
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._monitor = None
            self._monitor_lock = Lock()

        @contextmanager
        def tqdm(self, *args, **kwargs):
            """
            Opens a progress observer that keeps track of queue tasks.

            :param args: positional arguments for tqdm
            :param kwargs: key word arguments for tqdm
            :return: tqdm progress observer
            :rtype: tqdm
            """
            try:
                with self._monitor_lock:
                    self._monitor = tqdm(*args, **kwargs)
                yield self._monitor

            finally:
                with self._monitor_lock:
                    self._monitor.close()
                    self._monitor = None

        @wraps(qclass.task_done)
        def task_done(self):
            with self._monitor_lock:
                if self._monitor is not None:
                    self._monitor.update(1)

            super().task_done()

    _MonitoredQueue.__name__ = f'Monitored{qclass.__name__}'
    _MonitoredQueue.__doc__ = qclass.__doc__

    return _MonitoredQueue


@monitor_queue
class MonitoredQueue(queue.Queue):
    """monitored queue"""


class DelayQueue(queue.PriorityQueue):
    """
    A FIFO queue that ensures each item remains for a minimum amount of time in it.
    """
    def __init__(self, delay=0., maxsize=0):
        """
        :param delay: minimum time each item has to remain in queue
        :type delay: float
        :param maxsize: maximum size for queue
        :type maxsize: int
        """
        super().__init__(maxsize)

        assert delay >= 0, 'delay must not be negative'
        self._delay = delay

    @property
    def delay(self):
        return self._delay

    @wraps(queue.PriorityQueue.put)
    def put(self, item, *args, **kwargs):
        super().put((time(), item), *args, **kwargs)

    @wraps(queue.PriorityQueue.get)
    def get(self, *args, **kwargs):
        timestamp, item = super().get(*args, **kwargs)

        sleep(max(0, evaluate(self._delay) - (time() - timestamp)))

        return item


@monitor_queue
class MonitoredDelayQueue(DelayQueue):
    """monitored delay queue"""


class SingletonMeta(type):
    """
    This meta class ensures that there exists only one instance of its instances at a time. It's
    safe to instantiate it multiple times.
    """
    @wraps(type.__init__)
    def __init__(cls, name, bases, dct):
        super(SingletonMeta, cls).__init__(name, bases, dct)

        _new = cls.__new__

        @wraps(cls.__new__)
        def _new_wrapper(kls, *_, **__):
            if kls.instance is None:
                kls.instance = _new(kls)
            return kls.instance

        cls.instance = None
        cls.__new__ = classmethod(_new_wrapper)


class SingletonABCMeta(SingletonMeta, abc.ABCMeta):
    """Mixin for `abc.ABCMeta` and `SingletonMeta`"""


class SingletonContextABC(metaclass=SingletonABCMeta):
    """
    This class combines two popular programming patterns: singleton and context manager. The
    singleton property ensures that there's always one instance at most of this class unless you
    break that mechanism intentionally by calling :meth:`detach()`.
    As the context may be entered multiple times e.g. by nesting `with` statements or accessing that
    instance from multiple threads it keeps track of how often it was is_open. When the context is
    entered for the first and left for the last time a corresponding callback method is toggled.

    In order to implement the singleton context protocol you may overwrite the following callback
    methods:

      * :meth:`__on__open__(self) -> None`
      * :meth:`__on__close__(self) -> None`

    """
    __ctx_lock = RLock()
    __ctx_counter = 0

    def __new__(cls):
        """
        To prevent instantiation of this class without having abstract methods we perform a manual
        check.

        :return: new instance
        :rtype: SingletonContextABC
        """
        if cls is SingletonContextABC:
            raise TypeError(f'Can\'t instantiate abstract class {cls.__name__}.')

        return object.__new__(cls)

    def __setattr__(self, key, value):
        """
        The object can be modified as long as no context is open. Once that's the case calling this
        method won't have an effect anymore and fails silently.

        :param key: attribute name
        :type key: str
        :param value: new attribute value
        """
        with type(self).__ctx_lock:
            if not self.active():
                super().__setattr__(key, value)

            else:
                logger.debug(f'try to modify "{key}" of is_open singleton context')

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @classmethod
    def is_open(cls):
        """
        Counts how often the context was opened and not yet closed.

        :return: Value of internal context counter.
        :rtype: int
        """
        return cls.__ctx_counter

    @classmethod
    def active(cls):
        """
        Check whether any context is active

        :return: true if any context is active
        :rtype: bool
        """
        return cls.is_open() != 0

    @classmethod
    def detach(cls):
        """
        As this class implements the singleton pattern it holds a reference to its only instance
        internally. Also it has a counter that tracks how often the context is is_open currently.
        Both are reset by this method and it may happen that multiple instances of this class exist.

        .. warning:: Calling this method may have unexpected side effects!
        """
        if cls.instance:
            logger.debug(f'Detaching {cls.instance}. This may have unexpected side effects!')

            with cls.__ctx_lock:
                cls.__ctx_counter = 0

                del cls.instance
                cls.instance = None

    def __on_open__(self):
        """
        Callback method that is triggered when the context is entered for the first time or was
        closed before.
        """
        pass

    def __on_close__(self):
        """Callback method that is triggered when the context is left for the last time."""
        pass

    def open(self):
        """Opens the singleton context."""
        cls = type(self)

        with cls.__ctx_lock:
            if cls.__ctx_counter == 0:
                self.__on_open__()

            cls.__ctx_counter += 1

    def close(self):
        """Closes the singleton context."""
        cls = type(self)

        with cls.__ctx_lock:
            # This check is necessary for cases where a thread crashes during the call of
            # `__on_open__` and therefore enters this method without incrementing the context
            # counter before.
            if cls.__ctx_counter > 0:
                cls.__ctx_counter -= 1

                if cls.__ctx_counter == 0:
                    self.__on_close__()



# Standard library modules.

# Third party modules.

# Local modules
from mosquito.utils import (
    MosquitoError,
    monitor_queue,
    MonitoredQueue,
    DelayQueue,
    MonitoredDelayQueue
)
from mosquito.api import (
    available_attributes,
    register_attributes,
    attribute,
    swarm,
    observer
)
from mosquito.swarm import Swarm
from mosquito.scheduler import Scheduler

# Globals and constants variables.
MOSQUITO = mosquito = r"""
mosquito        \             /
                 \     |     /
                 /   \ | /   \
                 \    \|/    /
                  \,  o^o  ,/
                    \,/"\,/
            ,,,,----,{/X\},----,,,,
   ,,---''''      _-'{\X/}'-_      ''''---,,
 /'            ,-'/   \V/   \'-,            '\
(        ,--''/   |   (_)   |   \''--,        )
 '--,,-''    |    |   /_\   |   |     ''-,,--'
            /'    |  (_-_)  |   '\
           /     /'   \_/   '\    \
          /     /     (_)     \    \
               /       V       \
              /                 \
             /                   \
""".strip()



# Standard library modules.
from queue import Empty
from threading import Lock
from itertools import count
from functools import wraps
from collections import Counter
from contextlib import contextmanager

# Third party modules.
import requests

# Local modules
from mosquito.utils import (
    logger, args_to_kwargs, evaluate, MosquitoError, DelayQueue, SingletonContextABC)


# Globals and constants variables.


class Session(requests.Session):
    """
    Monitorable session object
    """
    __id_counter = count()

    def __init__(self, **kwargs):
        """
        :param kwargs: attributes to be set (see :meth:`update`)
        """
        super().__init__()

        self._id = next(self.__id_counter)
        self._observer = SessionObserver()
        self._observers = [self.observer]

        self.update(kwargs)

    def __str__(self):
        return f'{self.__class__.__name__}(id={self.id})'

    @property
    def id(self):
        return self._id

    @property
    def observer(self):
        return self._observer

    @args_to_kwargs(ftype='method')
    def update(self, **kwargs):
        """
        Patch attributes of a session. Dict like attributes will be updated. Other attributes are
        casted if possible and then set.

        :param kwargs: attributes to be updated
        """
        for name, value in kwargs.items():
            if name not in self.__attrs__:
                raise AttributeError(
                    f'\'{self.__class__.__name__}\' object has no attribute \'{name}\''
                )

            attr_t = type(getattr(self, name, None))

            # update dict like attributes (dict, cookie jar, ...)
            if hasattr(attr_t, 'update'):
                getattr(self, name).update(value)
                continue

            # set (casted) value
            try:
                setattr(self, name, attr_t(value))

            except TypeError:
                setattr(self, name, value)

    def register_observer(self, observer):
        """
        Attach observer to session.

        :param observer: a session observer instance
        :type observer: SessionObserver
        """
        self._observers.append(observer)

    @wraps(requests.Session.request)
    def request(self, method, url, *args, **kwargs):
        logger.debug(f'{self}: {method} {url}')
        response = super().request(method, url, *args, **kwargs)

        for observer in self._observers:
            observer.update(self, response)

        return response


class SessionObserver:
    """An observer for sessions that collects request statistics."""
    def __init__(self):
        self._lock = Lock()

        self.count = None
        self.duration = None
        self.sessions = None
        self.status_codes = None

        self.reset()

    def __str__(self):
        mean_duration = self.duration / (self.count or 1)
        return \
            f'{self.__class__.__name__}(' \
            f'count={self.count}, ' \
            f'mean_duration={mean_duration:.5f}s, ' \
            f'sessions={self.sessions}, ' \
            f'status_codes={self.status_codes}' \
            f')'

    def reset(self):
        """Reset session observer."""
        with self._lock:
            self.count = 0
            self.duration = 0
            self.sessions = Counter()
            self.status_codes = Counter()

    def update(self, session, response):
        """
        Update session statistics.

        :param session: session object
        :type session: Session
        :param response: response object
        :type response: requests.Response
        """
        with self._lock:
            self.count += 1
            self.duration += response.elapsed.total_seconds()
            self.sessions.update([session.id])
            self.status_codes.update([response.status_code])


class IdentityFactory:
    """Generate session identities from given attributes."""
    def __init__(self, attributes, require=None):
        """
        :param attributes: attributes to generate identities from, usually a dictionary where keys
                           are attribute names and values that evaluate to iterables with actual
                           attribute values
        :param require: container with keys of mandatory attributes
        """
        self._attributes = attributes
        self._require = require

    def __iter__(self):
        # evaluate all attributes
        attributes_dict = {attr: evaluate(value) for attr, value in self._attributes.items()} or {}

        # validate attributes
        for attribute in attributes_dict.keys():
            if attribute not in Session.__attrs__:
                raise AttributeError(f'unknown attribute "{attribute}"')

        # validate requirements
        for attribute in evaluate(self._require) or ():
            if attribute not in Session.__attrs__:
                raise AttributeError(f'unknown attribute "{attribute}"')

            if attribute not in self._attributes:
                raise MosquitoError(f'requirement "{attribute}" is not fulfilled')

        # generate identities
        if attributes_dict:
            keys, values = zip(*attributes_dict.items())

            for i, vals in enumerate(zip(*values)):
                yield dict(zip(keys, vals))

        else:
            yield {}


class SessionFactory(IdentityFactory):
    """Generate request sessions from given attributes."""
    def __iter__(self):
        for attributes in super().__iter__():
            yield Session(**attributes)


class Swarm(SingletonContextABC):
    """
    A swarm is a collection of request sessions that manages initialisation, teardown and access to
    those. Note that there exists only one swarm at a time and manipulation of the swarm object only
    has effects if no swarm context is active.
    """
    __attrs__ = ['on_init', 'delay', 'require', *Session.__attrs__]
    observer = SessionObserver()

    def __init__(self, delay=0., on_init=None, require=None, **session_attributes):
        """
        :param delay: delay for each session that determines max request frequency
        :type delay: float
        :param on_init: call back function to initialize sessions
        :param require: a list of attributes that have to be provided to each session
        :param session_attributes: session attributes
        """
        self._delay = delay
        self._on_init = on_init
        self._session_factory = SessionFactory(session_attributes, require)
        self._lock = Lock()
        self._sessions = None
        self._size = 0

    def __len__(self):
        """Counts items in session queue."""
        with self._lock:
            return self._size

    @contextmanager
    def session(self, timeout=None):
        """
        A context that provides the next available session object. If time passed since the last
        time using that session is less than :attr:`delay` this method blocks.

        :param timeout: max time to wait for session
        :return: session
        :rtype: Session
        """
        if self._sessions is None:
            raise MosquitoError('sessions are available in open swarm context only')

        session = self._sessions.get(timeout)

        try:
            yield session

        finally:
            self._sessions.put(session)

    def __on_open__(self):
        """Initialize sessions."""
        self._sessions = DelayQueue(delay=evaluate(self._delay))

        for session in self._session_factory:
            session.register_observer(self.observer)

            if self._on_init:
                self._on_init(session)

            self._size += 1
            self._sessions.put(session)

        logger.debug(f'opened swarm with {len(self)} sessions')

    def __on_close__(self):
        """Close all sessions."""
        try:
            while True:
                session = self._sessions.get_nowait()
                session.close()

        except Empty:
            pass

        self._size = 0
        self._sessions = None
        logger.debug(f'closed swarm successfully')



# Standard library modules.
import os
import logging
from functools import wraps

# Third party modules.

# Local modules

# Globals and constants variables.
HTTPBIN_BASE_URL = os.environ.get('HTTPBIN_BASE_URL')

logger = logging.getLogger('mosquito')

if not HTTPBIN_BASE_URL:
    HTTPBIN_BASE_URL = 'https://httpbin.org'

    logger.info(f'Fell back on {HTTPBIN_BASE_URL} for testing. Consider to run a local instance '
                f'of httpbin to increase testing performance.')


def httpbin(path=''):
    if not path.startswith('/'):
        path = f'/{path}'

    return f'{HTTPBIN_BASE_URL}{path}'


def k_of_n(k, n, message=None):
    """
    Decorator for test case methods that let's a test pass if at least k of n runs pass. This adds
    some tolerance to e.g. time critical test cases but will still catch systematical errors.

    :param k: number of successful attempts to pass
    :type k: int
    :param n: max attempts
    :type n: int
    :param message: custom message to be attached to exception
    :type message: str
    :return: decorator
    :rtype: function
    """
    assert 0 < k <= n

    def decorator(test_method):
        @wraps(test_method)
        def method_wrapper(*args, **kwargs):
            failures = 0

            for i in range(n):
                try:
                    test_method(*args, **kwargs)

                except AssertionError as error:
                    failures += 1

                    # fails if goal cannot be achieved anymore
                    if failures > n - k:
                        if message:
                            orig_message = f'{error.args[0]} -- ' if error.args else ''
                            error.args = (f'{orig_message}{message.strip()}', *error.args[1:])

                        raise error

                # aborts if goal is achieved
                if i - failures >= k:
                    break

        return method_wrapper

    return decorator


time_critical = k_of_n(
    k=2,
    n=3,
    message='Errors occuring while testing time critical components may be caused by slow '
            'execution e.g. due a high cpu load.'
)



#!/usr/bin/env python3
from functools import wraps

# Third party modules.

# Local modules
from mosquito.utils import MosquitoError
from mosquito.swarm import Session, Swarm

# Globals and constants variables.


class Scheduler:
    """
    Offers an interface similar to `requests.Session` but schedules all requests to a pool of
    sessions, the so called "swarm". Also it allows to be more obstinate and repeats requests on
    failure.
    """
    def __init__(self, swarm, max_attempts=1, repeat_on=None):
        """

        :param swarm: swarm instance
        :type swarm: Swarm
        :param max_attempts: maximum repetitions for requests on given errors
        :type max_attempts: int
        :param repeat_on: error codes where requests are repeated
        """
        self._swarm = swarm
        self._max_attempts = int(max_attempts)
        self._repeat_on = tuple(repeat_on) if repeat_on else ()

        assert self._max_attempts > 0, f'"max_attempts" has to be a positive integer.'
        assert 200 not in self._repeat_on, 'It does not make sense to repeat on status 200'

    swarm = property(fget=lambda self: self._swarm)
    max_attempts = property(fget=lambda self: self._max_attempts)
    repeat_on = property(fget=lambda self: tuple(self._repeat_on))

    def _request(self, request_method, url, **kwargs):
        for i in range(1, self._max_attempts + 1):
            with self._swarm.session() as session:
                response = request_method(session, url, **kwargs)

            status_code = response.status_code

            # repeat if status code is whitelisted
            if status_code in self._repeat_on:
                continue

            # return response on success
            elif status_code == 200:
                return response

            raise MosquitoError(f'request failed with code {status_code} on {i} attempt')

        raise MosquitoError(f'request failed after {self._max_attempts} attempts')

    @wraps(Session.get)
    def get(self, *args, **kwargs):
        return self._request(Session.get, *args, **kwargs)

    @wraps(Session.options)
    def options(self, *args, **kwargs):
        return self._request(Session.options, *args, **kwargs)

    @wraps(Session.head)
    def head(self, *args, **kwargs):
        return self._request(Session.head, *args, **kwargs)

    @wraps(Session.post)
    def post(self, *args, **kwargs):
        return self._request(Session.post, *args, **kwargs)

    @wraps(Session.put)
    def put(self, *args, **kwargs):
        return self._request(Session.put, *args, **kwargs)

    @wraps(Session.patch)
    def patch(self, *args, **kwargs):
        return self._request(Session.patch, *args, **kwargs)

    @wraps(Session.delete)
    def delete(self, *args, **kwargs):
        return self._request(Session.delete, *args, **kwargs)



#!/usr/bin/env python3
# Standard library modules.
import os
from unittest import TestLoader, TextTestRunner

# Third party modules.

# Local modules
from mosquito_test import httpbin

# Globals and constants variables.


def mosquito_test():
    path = os.path.dirname(__file__)

    print(f'Use httpbin instance on {httpbin()} for testing.')
    print(f'Run unit tests in: {path}\n')

    loader = TestLoader()
    suite = loader.discover(start_dir=path)
    runner = TextTestRunner()

    runner.run(suite)


if __name__ == '__main__':
    mosquito_test()



# Standard library modules.
from unittest import TestCase

# Third party modules.

# Local modules
from mosquito import MosquitoError
from mosquito.api import registry, register_attributes, attribute, swarm
from mosquito_test.utils import httpbin

# Globals and constants variables.


class TestFunctions(TestCase):
    def setUp(self):
        self.attribute_registry_backup = dict(registry())

    def tearDown(self):
        registry().clear()
        registry().update(self.attribute_registry_backup)

    def test_register_attributes(self):
        register_attributes(headers=[42])
        self.assertEqual([42], registry().get('headers'))

        with self.assertRaises(AttributeError):
            register_attributes(fnord=[])

    def test_attribute(self):
        attribute('headers')([42])
        self.assertEqual([42], registry().get('headers'))

        with self.assertRaises(AttributeError):
            attribute('fnord')([])

    def test_swarm(self):
        user_agents = {'foo', 'bar', 'baz'}
        register_attributes(headers=[{'user-agent': ua} for ua in user_agents])

        with swarm(max_attempts=2, repeat_on=(503,)) as scheduler:
            self.assertEqual(3, len(scheduler.swarm))

            uas = set()
            for _ in range(len(user_agents)):
                response = scheduler.get(httpbin('/user-agent'))
                uas.add(response.json().get('user-agent'))

            self.assertEqual(user_agents, uas)

            with self.assertRaises(MosquitoError):
                _ = scheduler.get(httpbin('/status/503'))



# Standard library modules.
from contextlib import contextmanager

# Third party modules.

# Local modules
from mosquito.swarm import Swarm
from mosquito.scheduler import Scheduler

# Globals and constants variables.
ATTRIBUTE_REGISTRY = dict()


def registry():
    """
    :return: a reference on the global attribute registry
    :rtype: dict
    """
    return ATTRIBUTE_REGISTRY


def available_attributes():
    """
    :return: names of all available attributes
    :rtype: list[str]
    """
    return sorted(getattr(Swarm, '__attrs__'))


def register_attributes(**kwargs):
    """
    Update global attribute registry.

    :param kwargs: attribute key values mapping
    """
    for attr, value in kwargs.items():
        if attr not in available_attributes():
            raise AttributeError(f'unknown attribute "{attr}"')

        ATTRIBUTE_REGISTRY[attr] = value


def attribute(attr):
    """
    Decorator for an attribute callback.

    :param attr: name of the attribute to be registered
    :return: decorator to register attribute
    """
    def attribute_decorator(value):
        register_attributes(**{attr: value})

        return value

    return attribute_decorator


@contextmanager
def swarm(*args, **kwargs):
    """
    Context that sets up a session swarm using registered attributes and provides a preconfigured
    scheduler object.

    :param args: positional arguments for :class:`Scheduler`
    :param kwargs: key word arguments for :class:`Scheduler`
    :return: scheduler
    :rtype: Scheduler
    """
    _swarm = Swarm(**registry())

    try:
        _swarm.open()
        yield Scheduler(_swarm, *args, **kwargs)

    finally:
        _swarm.close()


def observer():
    return Swarm.observer



# Standard library modules.
from unittest import TestCase

# Third party modules.

# Local modules
from mosquito import MosquitoError, Swarm, Scheduler
from mosquito_test.utils import httpbin

# Globals and constants variables.


class TestScheduler(TestCase):
    def setUp(self):
        Swarm.detach()

        self.swarm = Swarm()
        self.swarm.open()

        self.scheduler = Scheduler(self.swarm)

    def tearDown(self):
        self.swarm.close()

    def test__request(self):
        with self.assertRaises(MosquitoError):
            self.scheduler.get(httpbin('/status/404'))

        scheduler = Scheduler(self.swarm, repeat_on=(503,))
        try:
            scheduler.get(httpbin('/status/503'))
        except MosquitoError as me:
            self.assertIn('1', str(me))

        scheduler = Scheduler(self.swarm, max_attempts=3, repeat_on=(503,))
        try:
            scheduler.get(httpbin('/status/503'))
        except MosquitoError as me:
            self.assertIn('3', str(me))

        with self.assertRaises(AssertionError):
            _ = Scheduler(self.swarm, max_attempts=0)

        with self.assertRaises(AssertionError):
            _ = Scheduler(self.swarm, repeat_on=(200,))

    def test_get(self):
        response = self.scheduler.get(httpbin('/get'))
        self.assertEqual('GET', response.request.method)

    def test_options(self):
        response = self.scheduler.options(httpbin('/get'))
        self.assertEqual('OPTIONS', response.request.method)

    def test_head(self):
        response = self.scheduler.head(httpbin('/get'))
        self.assertEqual('HEAD', response.request.method)

    def test_post(self):
        response = self.scheduler.post(httpbin('/post'))
        self.assertEqual('POST', response.request.method)

    def test_put(self):
        response = self.scheduler.put(httpbin('/put'))
        self.assertEqual('PUT', response.request.method)

    def test_patch(self):
        response = self.scheduler.patch(httpbin('/patch'))
        self.assertEqual('PATCH', response.request.method)

    def test_delete(self):
        response = self.scheduler.delete(httpbin('/delete'))
        self.assertEqual('DELETE', response.request.method)



#!/usr/bin/env python3
# Standard library modules.
import queue
from time import time, sleep
from unittest import TestCase
from statistics import median

# Third party modules.

# Local modules
from mosquito.utils import (
    args_to_kwargs, evaluate, monitor_queue, MonitoredQueue, DelayQueue, SingletonMeta,
    SingletonContextABC
)
from mosquito_test.utils import k_of_n, time_critical

# Globals and constants variables.


class TestFunctions(TestCase):
    def test_k_of_n(self):
        # check logical correctness
        for k, n in [(2, 3), (2, 4), (3, 5)]:
            for i in range(2**n):
                # integer to bit tuple (big endian)
                ibit = tuple((i >> j) & 1 for j in range(n))

                states = iter(ibit)

                @k_of_n(k, n)
                def test_mock():
                    assert next(states)

                if sum(ibit) >= k:
                    test_mock()

                else:
                    with self.assertRaises(AssertionError):
                        test_mock()

        # check if message wrapping works
        @k_of_n(1, 1, 'fnord')
        def test_mock(msg=None):
            if msg:
                assert False, msg

            assert False

        # custom message only
        try:
            test_mock()
            self.fail()

        except AssertionError as error:
            self.assertNotIn('--', str(error))
            self.assertIn('fnord', str(error))

        # merging original and custom message
        try:
            test_mock('foo')
            self.fail()

        except AssertionError as error:
            self.assertIn('foo --', str(error))
            self.assertIn('fnord', str(error))

    def test_args_as_kwargs(self):
        @args_to_kwargs
        def mock(*args, **kwargs):
            return args, kwargs

        class Mock:
            @args_to_kwargs
            def static_1(*args, **kwargs):
                return args, kwargs

            @staticmethod
            @args_to_kwargs(ftype='static')
            def static_2(*args, **kwargs):
                return args, kwargs

            @args_to_kwargs(ftype='method')
            def method(self, *args, **kwargs):
                return args, kwargs

            @classmethod
            @args_to_kwargs(ftype='classmethod')
            def classmethod(cls, *args, **kwargs):
                return args, kwargs

            @args_to_kwargs(ftype='method', keep=3)
            def keep(self, a, b, c, *args, **kwargs):
                return a, b, c, args, kwargs

            @args_to_kwargs
            def fail(self, *args, **kwargs):
                return args, kwargs

        with self.assertRaises(AssertionError):
            @args_to_kwargs(ftype='foo')
            def foo():
                pass

        funcs = (mock, Mock.static_1, Mock.static_2, Mock.method, Mock.classmethod)
        args = ((), (), (), (Mock(),), ())

        for func, arg in zip(funcs, args):
            args, kwargs = func(*arg)
            self.assertEqual((), args)
            self.assertEqual({}, kwargs)

            # key word arguments are kept
            args, kwargs = func(*arg, foo=42)
            self.assertEqual((), args)
            self.assertEqual(dict(foo=42), kwargs)

            # positional argument is converted to key word arguments
            args, kwargs = func(*arg, dict(foo=42))
            self.assertEqual((), args)
            self.assertEqual(dict(foo=42), kwargs)

            # key word arguments dominate positional arguments
            args, kwargs = func(*arg, dict(foo=1337), foo=42)
            self.assertEqual((), args)
            self.assertEqual(dict(foo=42), kwargs)

            # right most argument dominates
            args, kwargs = func(*arg, dict(foo=1337), dict(foo=42))
            self.assertEqual((), args)
            self.assertEqual(dict(foo=42), kwargs)

            # multiple positional arguments
            args, kwargs = func(*arg, dict(foo=13), dict(foo=37), foo=42)
            self.assertEqual((), args)
            self.assertEqual(dict(foo=42), kwargs)

            # it also works with mappings
            args, kwargs = func(*arg, (('foo', 42),))
            self.assertEqual((), args)
            self.assertEqual(dict(foo=42), kwargs)

            # error when positional argument is not a dict like
            with self.assertRaises(TypeError):
                func(*arg, None)

        # first three positional arguments are kept
        a, b, c, args, kwargs = Mock().keep(13, 37, 42, dict(foo=42), bar=1337)
        self.assertEqual(13, a)
        self.assertEqual(37, b)
        self.assertEqual(42, c)
        self.assertEqual((), args)
        self.assertEqual(dict(foo=42, bar=1337), kwargs)

        # wrong decorator parametrisation results in error
        with self.assertRaises(TypeError):
            Mock().fail()

    def test_evaluate(self):
        self.assertEqual(None, evaluate(None))
        self.assertTupleEqual((0, 1, 2), evaluate([0, 1, 2]))
        self.assertTupleEqual((0, 1, 2), evaluate(range(3)))
        self.assertTupleEqual((0, 1, 2), evaluate(lambda: [0, 1, 2]))
        self.assertTupleEqual((0, 1, 2), evaluate(lambda: range(3)))


class TestMonitorQueue(TestCase):
    def setUp(self):
        self.mq = MonitoredQueue()

        for i in range(3):
            self.mq.put(f'item_{i}')

    def test_monitor_queue(self):
        queue_class = monitor_queue(queue.Queue)

        self.assertIn('__init__', queue_class.__dict__)
        self.assertIn('tqdm', queue_class.__dict__)
        self.assertIn('task_done', queue_class.__dict__)

    def test_tqdm(self):
        with self.mq.tqdm() as monitor:
            for i in range(2):
                self.mq.get_nowait()
                self.mq.task_done()

                self.assertEqual(i+1, monitor.n)

    def test_task_done(self):
        self.assertEqual(3, self.mq.unfinished_tasks)

        self.mq.get_nowait()
        self.mq.task_done()

        self.assertEqual(2, self.mq.unfinished_tasks)


class TestDelayQueue(TestCase):
    def test___init__(self):
        with self.assertRaises(AssertionError):
            _ = DelayQueue(delay=-1.)

    @time_critical
    def test_put_get_delay(self):
        n = 4
        t_unit = 3e-4
        delay = 5 * t_unit
        dq = DelayQueue(delay=delay)

        put_times = []
        get_times = []
        overheads = []

        for i in range(n):
            dq.put(i)
            put_times.append(time())

            sleep(t_unit)

        for i in range(n):
            j = dq.get()
            get_times.append(time())

            self.assertEqual(i, j)

        for put_time, get_time in zip(put_times, get_times):
            overheads.append(get_time - put_time - delay)
            self.assertLess(put_time, get_time)

        self.assertLess(median(overheads), 5e-4)


class TestSingletonABCMeta(TestCase):
    class SingletonMockA(metaclass=SingletonMeta):
        def __init__(self):
            pass

    class SingletonMockB(metaclass=SingletonMeta):
        def __init__(self):
            pass

    def test(self):
        s_1 = self.SingletonMockA()
        s_2 = self.SingletonMockA()
        s_3 = self.SingletonMockB()
        s_4 = self.SingletonMockB()

        self.assertIs(s_1, s_2)
        self.assertIs(s_3, s_4)
        self.assertIsNot(s_1, s_3)


class TestSingletonContext(TestCase):
    class MockError(Exception):
        pass

    class SingletonContextMock(SingletonContextABC):
        def __init__(self, fail_on_open=None, fail_on_close=None):
            self.fail_on_open = fail_on_open
            self.fail_on_close = fail_on_close
            self.ran_on_open = False
            self.ran_on_close = False

        def __on_open__(self):
            if self.fail_on_open:
                raise self.fail_on_open('failed on open')

            self.ran_on_open = True

        def __on_close__(self):
            if self.fail_on_close:
                raise self.fail_on_close('failed on close')

            self.ran_on_close = True

    def setUp(self):
        self.SingletonContextMock.detach()

    def test__new__(self):
        with self.assertRaises(TypeError):
            _ = SingletonContextABC()

    def test_setattr(self):
        s = self.SingletonContextMock()

        s.foo = 42
        self.assertEqual(42, s.foo)

        s.foo = 1337
        self.assertEqual(1337, s.foo)

        with s:
            s.foo = 42
            self.assertEqual(1337, s.foo)

        s.foo = 42
        self.assertEqual(42, s.foo)

    def test_context(self):
        # test for errors occurring on startup / teardown
        setups = ({}, dict(fail_on_open=self.MockError), dict(fail_on_close=self.MockError))
        for kwargs in setups:
            sc = self.SingletonContextMock(**kwargs)

            try:
                with sc:
                    self.assertEqual(1, sc.is_open())

            except self.MockError:
                if not kwargs.values():
                    self.fail('did not expect any errors to occur')

            self.assertEqual(0, sc.is_open())

            self.SingletonContextMock.detach()

        # test for errors occurring in context
        sc = self.SingletonContextMock()
        with self.assertRaises(self.MockError):
            with sc:
                raise self.MockError()

        self.assertEqual(0, sc.is_open())

    def test_is_open(self):
        sc = self.SingletonContextMock()

        self.assertEqual(0, sc.is_open())

        with self.SingletonContextMock():
            self.assertEqual(1, sc.is_open())

            with self.SingletonContextMock():
                self.assertEqual(2, sc.is_open())

                with self.SingletonContextMock():
                    self.assertEqual(3, sc.is_open())

                self.assertEqual(2, sc.is_open())

            self.assertEqual(1, sc.is_open())

        self.assertEqual(0, sc.is_open())

    def test_active(self):
        sc = self.SingletonContextMock()

        self.assertFalse(sc.active())

        with self.SingletonContextMock():
            self.assertTrue(sc.active())

            with self.SingletonContextMock():
                self.assertTrue(sc.active())

            self.assertTrue(sc.active())

        self.assertFalse(sc.active())

    def test_detach(self):
        self.assertIsNone(self.SingletonContextMock.instance)

        with self.SingletonContextMock():
            pass

        self.assertIsNotNone(self.SingletonContextMock.instance)

        self.SingletonContextMock.detach()

        self.assertIsNone(self.SingletonContextMock.instance)

    def test_open_close_manual(self):
        sc = self.SingletonContextMock()

        self.assertFalse(sc.ran_on_open)
        self.assertFalse(sc.ran_on_close)

        sc.open()

        self.assertTrue(sc.ran_on_open)
        self.assertFalse(sc.ran_on_close)

        sc.close()

        self.assertTrue(sc.ran_on_open)
        self.assertTrue(sc.ran_on_close)

    def test_open_close_auto(self):
        sc = self.SingletonContextMock()

        self.assertFalse(sc.ran_on_open)
        self.assertFalse(sc.ran_on_close)

        with sc:
            pass

        self.assertTrue(sc.ran_on_open)
        self.assertTrue(sc.ran_on_close)



#!/usr/bin/env python3

# Standard library modules.
import os

# Third party modules.
from setuptools import setup, find_packages

# Local modules.

# Globals and constants variables.
BASEDIR = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(BASEDIR, 'README.md'), 'r') as fp:
    LONG_DESCRIPTION = fp.read()

PACKAGES = find_packages()

INSTALL_REQUIRES = ['requests', 'tqdm']
EXTRAS_REQUIRE = {'develop': ['nose']}

ENTRY_POINTS = {}

setup(
    name='mosquito',
    version='0.1.0',
    author="luoc",
    author_email="luoc@posteo.org",
    description="a request obfuscator and web scraping toolkit",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://luoc.eu",
    packages=PACKAGES,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    entry_points=ENTRY_POINTS,
    test_suite='nose.collector',
    scripts=[],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
)



# Standard library modules.
from time import time
from math import isclose
from itertools import repeat
from unittest import TestCase
from collections import Counter

# Third party modules.
import requests

# Local modules
from mosquito import MosquitoError
from mosquito.swarm import Session, SessionObserver, IdentityFactory, SessionFactory, Swarm
from mosquito_test.utils import httpbin, time_critical

# Globals and constants variables.


class TestSession(TestCase):
    def test_id(self):
        sessions = [Session() for _ in range(5)]

        for i, session in enumerate(sessions):
            self.assertEqual(i, session.id - sessions[0].id)

    def test_update(self):
        # 1. update attributes via kwargs
        # 2. update attributes via args
        # 3. update attributes via both, demonstrating that kwargs dominate
        # 4. update attributes via args using mapping
        args_l = [
            (),
            (dict(headers=dict(foo=42), max_redirects=42.0, auth=('foo', 'bar')),),
            (dict(headers=dict(foo=13), max_redirects=37.0, auth=('bar', 'foo')),),
            ((('headers', dict(foo=42)), ('max_redirects', 42.0), ('auth', ('foo', 'bar')),), )
        ]
        kwargs_l = [
            dict(headers=dict(foo=42), max_redirects=42.0, auth=('foo', 'bar')),
            {},
            dict(headers=dict(foo=42), max_redirects=42.0, auth=('foo', 'bar')),
            {}
        ]

        for args, kwargs in zip(args_l, kwargs_l):
            s = Session()
            s.update(*args, **kwargs)

            self.assertIn('foo', s.headers)
            self.assertIn('user-agent', s.headers)
            self.assertEqual(42, s.headers.get('foo'))

            self.assertIs(int, type(s.max_redirects))
            self.assertEqual(42, s.max_redirects)

            self.assertEqual(('foo', 'bar'), s.auth)

            with self.assertRaises(AttributeError):
                s.update(foo=42)

        s = Session()
        # invalid type of positional argument
        with self.assertRaises(TypeError):
            s.update((None,))

    def test_request(self):
        with Session() as session:
            observer = SessionObserver()
            session.register_observer(observer)

            _ = session.get(httpbin('/get'))

            self.assertEqual(1, observer.count)
            self.assertEqual(1, session.observer.count)


class TestSessionObserver(TestCase):
    @staticmethod
    def generate_response(status_code=200):
        response = requests.Response()

        response.status_code = status_code

        return response

    def test_update_reset(self):
        sessions = [Session() for _ in range(3)]
        sm = SessionObserver()

        for i in range(10):
            sm.update(sessions[i % 3], self.generate_response((i % 2 * 100 + 200)))

        self.assertEqual(10, sm.count)
        self.assertEqual(0., sm.duration)
        self.assertEqual([3, 3, 4], sorted(sm.sessions.values()))
        self.assertEqual(Counter({200: 5, 300: 5}), sm.status_codes)

        sm.reset()

        self.assertEqual(0, sm.count)
        self.assertEqual(0., sm.duration)
        self.assertEqual(Counter(), sm.sessions)
        self.assertEqual(Counter(), sm.status_codes)


class TestIdentityFactory(TestCase):
    def test_identities(self):
        val = lambda: [42]

        if_1 = IdentityFactory({})
        if_2 = IdentityFactory({attr: vals for attr, vals in zip(Session.__attrs__, repeat(val))})
        if_3 = IdentityFactory({attr: vals for attr, vals in zip(Session.__attrs__, repeat(val))},
                               require=Session.__attrs__)
        if_4 = IdentityFactory(dict(foo=42))
        if_5 = IdentityFactory({}, require=['fnord'])
        if_6 = IdentityFactory({}, require=['headers'])

        self.assertEqual([{}], list(if_1))

        for im in (if_2, if_3):
            self.assertEqual(set(Session.__attrs__), set(list(im)[0].keys()))
            self.assertEqual({42}, set(list(im)[0].values()))

        with self.assertRaises(AttributeError):
            _ = list(if_4)

        with self.assertRaises(AttributeError):
            _ = list(if_5)

        with self.assertRaises(MosquitoError):
            _ = list(if_6)


class TestSessionFactory(TestCase):
    def setUp(self):
        self.factory_class = SessionFactory

    def test_sessions(self):
        attributes = dict(params=(dict(foo=i) for i in range(3)))
        sessions = tuple(SessionFactory(attributes))

        self.assertEqual(3, len(sessions))
        for i, session in enumerate(sessions):
            self.assertEqual(dict(foo=i), session.params)
            self.assertIsInstance(session, Session)


class TestSwarm(TestCase):
    def setUp(self):
        Swarm.detach()

    @time_critical
    def test_session_context(self):
        delay = .001
        iters = 32

        with Swarm(delay=delay, headers=[{}] * 4) as swarm:
            t_start = time()

            for i in range(iters):
                with swarm.session():
                    pass

            t_total = time() - t_start

            mean_duration = (t_total / iters) * len(swarm)

            # The above formula is correct only if iters is a multiple of swarm size or infinite
            self.assertEqual(0, iters % len(swarm))

            # ensure delay is hard lower limit
            self.assertLess(delay, mean_duration)

            # tolerance of 25%
            self.assertTrue(isclose(delay, mean_duration, rel_tol=25e-2))

    def test_on_init(self):
        sids = set()

        def init_callback(session):
            sids.add(session.id)

        with Swarm(on_init=init_callback, headers=[{}] * 3):
            pass

        self.assertEqual(3, len(sids))

    def test_open_close(self):
        swarm = Swarm(headers=[{}] * 3)

        self.assertEqual(0, len(swarm))
        with self.assertRaises(MosquitoError):
            with swarm.session():
                pass

        swarm.open()
        with swarm.session():
            pass

        self.assertEqual(3, len(swarm))

        swarm.close()

        self.assertEqual(0, len(swarm))
        with self.assertRaises(MosquitoError):
            with swarm.session():
                pass

        with swarm:
            self.assertEqual(3, len(swarm))
            with swarm.session():
                pass

        self.assertEqual(0, len(swarm))

        with Swarm() as swarm:
            self.assertEqual(1, len(swarm))
